#ifndef  charSetTranslator_h_included
#define  charSetTranslator_h_included

#include <utility>
#include <list>
#include <string>

namespace cxxtls
{

class CharSetTranslator
    ///  An object that performs mappings between ASCII character set ranges.
    ///  To use a CharSetTranslator, you construct it and then use either one of
    ///  of the following members:
    ///
    ///  @li translate()  -- translates whole strings
    ///  @li operator[]() -- translates individual characters
    ///
    ///  A translator is constructed using a pair of std::string which define
    ///  characters to be translated.  Sequences of characters may be specified
    ///  to reduce typing.  Here are some examples:
    ///
    ///@code
    ///
    ///  CharSetTranslator("z","z");
    ///
    ///    -- empty translation.  It leaves all input characters untouched.
    ///  
    ///  CharSetTranslator("a-z", "A-Z");
    ///
    ///    -- performs the upper case operation.  Only characters in the range
    ///       a-z will be affected.
    ///
    ///   
    ///  CharSetTranslator("\o000-\o377", " ");
    ///
    ///    -- converts all characters into space. When the output set is shorter
    ///       than the input set, the final character of the outputset is used for
    ///       any characters missed in the input set.
    ///
    ///  CharSetTranslator("a-fA-F", "1-61-6");
    ///
    ///    -- converts all letters to numbers.
    ///
    ///
    ///
    ///
    ///@endcode
{   // CharSetTranslator
    char  mapping_[256];  ///<  A mapping between an input character set and an output
                          ///   character set.

    bool  ok_;            ///<  Flag indicating that all is good with this container.

    char const *error_;   ///<  error text associated with ok_ being false

    struct LhsSubrange
        ///  Information needed to describe character ranges in the left hand side
        ///  of the mapping
    {
        int offset_;         // the character in the left hand side
        int mappedOffset_;   // the index position in the right hand side

        LhsSubrange(int lhsChar, int mappedRhsPosition)
        : offset_(lhsChar)
        , mappedOffset_(mappedRhsPosition)
        {
        }

    };

    typedef std::list< LhsSubrange > LhsRange;

public:

    bool ok() const { return ok_; }               ///< @return true to indicate a successful construction
    char const *error() const { return error_; }  ///< @return error message if any

    CharSetTranslator(std::string const &inset, std::string const &outset)
        ///  Create the mapping between the the characters defined in the inset and the outset and
        ///  store it in the mapping_ member.
        ///
        ///  Both the inset and the outset are specified in the style of regular
        ///  expression character sets:
        ///
        ///@code
        ///   set         := {char | charRange}
        ///   charRange   := char "-" char
        ///@endcode
        ///
        /// @note  to specify "-" as a single character, it must be first in the
        ///        set.  For example:   "-abc", includes "-" as a character, but
        ///        "a-z", specifies all characters in the range "a" thorugh "z".
        ///
        ///  For example:
        ///
        ///@code
        ///
        ///   -+a-fA-F0-9
        ///
        ///@endcode
        ///
        ///  Which means all of the following:
        ///  @li the character -
        ///  @li the character +
        ///  @li the range of characters 'a' through, and including 'f'
        ///  @li the range of characters 'A' through, and including 'F'
        ///  @li the range of characters '0' thorugh, and including '9' 
    {

        ok_    = true;
        error_ = "";

        LhsRange lhs;

        char rhs[256];

        populateLhs(unEscape(inset).c_str(), lhs);
        populateRhs(unEscape(outset).c_str(), rhs);
        populateMapping(lhs, rhs);

    }

    char operator[] (unsigned char i) const { return mapping_[i]; }  ///< map a character 


    std::string translate(std::string const &s) const
        /// Copy and translate a string.  If the construction of
        /// *this failed, then return empty string.
        ;

private:

    void populateLhs(char const *inset, LhsRange &lhs)
        /// Parse the left hand side of the character translation into a list
        /// of sub-ranges to be extracted from the right hand side.
        ;

    void populateRhs(char const *set, char *rhs)
        /// Parse a sequence of characters into a buffer of up to 256 character positions.
        /// The final character in the sequence will be used to populate the empty
        /// slots.
        ;

    void populateMapping(LhsRange const &lhs, char const *rhs)
        /// Combine the list of character specified on the left hand side with
        /// the values defined on the right to produce the mappingSet_ member.
        ;

public:

    static std::string unEscape(std::string const &in)
        ///  Convert an input string to an output string by translating the standard
        ///  C++ language escape character sequences into the characters they 
        ///  represent.  For example:
        ///@code
        ///
        ///   \n - newline
        ///   \r - return characters
        ///   \t - tab
        ///@endcode
        ///
        /// etc.
        ;


};  // CharSetTranslator

};  // namespace cxxlts


#endif
