#ifndef cpp_token_h_included
#define cpp_token_h_included


//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.




///@file cxxtls/cpp_token_stream.h 
///
///  This file defines the C++ lexical analysis tools.  It defines
///  the concept of a C++ 'token', also a template which parses tokens from
///  character sources (anything with an iterator), and a few concrete
///  instances of token streams.  The token sources defined below
///  use the lexer template, CPP_Token_Stream, to instantiate
///  CPP_Token_Source's.  That is, you can write your code using CPP_Tokens
///  to analyze the parsed source.  Instead of referring directly to one
///  of the concrete lexers (CPP_Stream_Token_Source, CPP_File_Token_Source,
///  or CPP_String_Token_Source) you can write your code using the abstract
///  lexer base class, CPP_Token_Source.  The virtual method call overhead is
///  likely to be small compared to the looping constructs typically found
///  inside the lexer (CPP_Token_Stream<>)
//


#include <fstream>
#include <string>
#include <portable_strstream.h>

namespace cxxtls
{


struct CPP_Token
  //
  /// An object that represents a C++ lexical token.  CPP_Token objects
  /// are usually obtained from token streams.  See CPP_Token_Stream's
  /// for an example use.
  ///
  /// The fastest way to compare a token to an expected value is to
  /// first verify that its 'type_' is correct.  Type is an integer so the
  /// comparison is quick.  Next, verify that that the text_ is correct
  /// If the type is wrong, there's no reason to check
  /// the string value.  To further speed up the comparison, if you want,
  /// you can verify that the first character of the text_ equals the
  /// desired string.  So here is some code that compares a token to some
  /// possible known values:
  ///
  /// @code
  ///    if(token.type_ == ';') return true;  // no text comparison is needed
  ///    if(token.type_ == CPP_Token::aln) return true; // any old identifier
  ///    if(token.type_ == CPP_Token::der) return true; // its a :: operator
  ///    if(token.type_ == CPP_Token::aln &&
  ///       token.text_[0] == 'i' && token.text_ == "if"
  ///      )  return true;  // it is the 'if' keyword
  /// @endcode
  //
{
  std::string text_;  ////< The string value of the token.  This will
		      ///   be the exact text parsed -- including any
		      ///   quotes, \'s etc.

  int         type_;  ///< An integer value definiting the token type.
		      ///  Single character tokens have a type value
		      ///  which equals their character value.
		      ///  Multi-character tokens have named values
		      ///  which are members of the token_types_
		      ///  nested enumeration.
		
  int         line_;  ///< what line does the token begin on

  std::string file_;  ///< what file was the token found in

  operator std::string const &() const
    //
    /// Automatic conversion to string.  All you get is the
    /// token's string text_ value.
  {
    return text_;
  }

  enum token_types_
    ///
    /// A list of names for the multi-character tokens
    /// and the end of file situation.
  {
     // one character operators have a value equal to their character value

		
     eof = 0x000, ///<    end of file
     eql = 0x100, ///<    ==
     neq = 0x101, ///<    !=
     leq = 0x102, ///<    \<=
     geq = 0x103, ///<    \>=
     sle = 0x104, ///<    \<<=
     sre = 0x105, ///<    \>>=
     dve = 0x106, ///<    /=
     ple = 0x107, ///<    +=
     mne = 0x108, ///<    -=
     tme = 0x109, ///<    *=
     mde = 0x10a, ///<    %=
     adr = 0x10b, ///<    &  the address of operator, not and operator
     orr = 0x10c, ///<    ||
     ore = 0x10d, ///<    |=
     ade = 0x10e, ///<    &=
     inc = 0x10f, ///<    ++
     dec = 0x110, ///<    --
     shr = 0x111, ///<    \>>
     shl = 0x112, ///<    \<<
     lan = 0x113, ///<    &&
     der = 0x114, ///<    ::
     teq = 0x115, ///<    ~=
     ceq = 0x116, ///<    ^=
     eli = 0x117, ///<    ...
     ptr = 0x118, ///<    ->
     nmb = 0x201, ///<    a number
     aln = 0x202, ///<    alpha numeric
     inv = 0x300, ///<    an invalid operator or other token
     special_tokens
  };

  static std::string type_name(int type);
    ///<
    /// static method giving a string representation of a
    /// token's type.
    ///
    /// @param type specifies the token type whose name should be returned


};

inline
std::ostream&
operator<<( std::ostream& o, CPP_Token const &r )
  /// output a token to a stream
{

  return o
	 << "\"" << r.file_
	 << "\", line " << r.line_
	 << " token, '" << r.text_ << "', is of type "
	 << CPP_Token::type_name(r.type_);

}

class CPP_Buffer_Token_Source; // see bottom of file

class CPP_Token_Stream_Prep
  //
  /// A CPP_Token_Stream, does not
  /// include preprocessor directives.  Doing so would
  /// greatly complicate any grammer.  Instead,
  /// preprocessor directives are handled only if
  /// desired in an out of band style.  To handle
  /// preprocessor directives, derive a class from
  /// this one (CPP_Token_Stream_Prep) and implement a
  /// virtual method which will be invoked by the stream
  /// as preprocessor directives are encountered.  Note that
  /// the '#' and any spaces surrounding it will not be
  /// included in the text passed to the virtual method.
  //
{
public:

   CPP_Token_Stream_Prep()
     /// default constructor
   {
   }

   virtual ~CPP_Token_Stream_Prep() {} //< destructor

   virtual void operator() (std::string const &text,
			    std::string const &file,
			    int                line) const = 0;
   ///<
   /// User defined callback handler to deal with preprocessor tokens.
   /// @param text  the text of the preprocessor directive
   /// @param file  the name of the file
   /// @param line     the line number of the preprocessor directive


};


template<class Iterator>
class CPP_Token_Stream
  //
  /// C++ token stream template (ie a source of CPP_Token objects)
  //
  /// A CPP_Token_Stream<Iterator> is a sequence of CPP_Token objects.
  /// A CPP_Token_Stream is a functor object.  Use its function call
  /// operator to get the next CPP_Token out of the stream.
  ///
  /// This class is a template, whose parameter is the kind of iterator
  /// to use when fetching characters out of the real stream.  You can
  /// use any of:
  ///   char*
  ///   std::string::const_iterator
  ///   std::istreambuf_iterator
  ///   other iterators which dereference to char
  ///
  /// Here is an example use:
  ///
  /// @code
  ///
  ///     #include <cxxtls/cpp_token_stream.h>
  ///
  ///     std::istreambuf_iterator<char>  cur(std::cin);
  ///     std::istreambuf_iterator<char>  end;
  ///
  ///     CPP_Token_Stream< std::istreambuf_iterator<char> >  ts(cur,end);
  ///
  ///     CPP_Token token_one;  ts(token_one); // declare and read tokens
  ///     CPP_Token token_two;  ts(token_two);
  ///
  /// @endcode
  ///
  /// When end of stream has occurred, the returned token will have a type
  /// which is CPP_Token::eof.  Multiple attempts to fetch the next token
  /// will all return this type without performing any subsequent parsing.
  //
  /// The only language features recognized by this lexer are:
  ///   the C++ token set
  ///   #line directives which are properly interpreted
  ///   all other # directives are removed from the source and ignored
  //
{
  Iterator                     cur_;                   ///<  current position in the stream
  Iterator                     end_;                   ///<  end of stream position
  int                          line_;                  ///<  line number in the stream
  std::string                  file_;                  ///<  name of current file in the stream
  bool                         initialized_;           ///<  has char_type_ been initilaized
  CPP_Token_Stream_Prep const *preprocessor_handler_;  ///<  handler for pre-processor directives
  char                         char_type_[256];        ///<  equivalent of ctype()
  friend class                 CPP_Buffer_Token_Source;

public:


  CPP_Token_Stream(Iterator const& next,
		   Iterator const& end,
		   std::string filename="",
		   CPP_Token_Stream_Prep const *handler=0
		  )
  :
    cur_(next),
    end_(end),
    line_(1),
    file_(filename),
    initialized_(0),
    preprocessor_handler_(handler)
  {
    ///<
    ///   Construct a token stream from a pair of iterators -- with
    ///   optional settings to define the name of the file being parsed
    ///   and a handler object that handles preprocessor directives.
  }


  void operator() (CPP_Token &token)
    /// Read one token out of the stream defined by a pair of iterators.
    /// Do not read past the end of the token stream.  When end of stream
    /// has occurred, the returned token type will be CPP_Token::eof
    ///
    /// @param token is the place to store the output token.

  {

    token.text_.resize(0);
    token.type_ = CPP_Token::inv;

    bool first_char = !initialized_;

    if(!initialized_)
    {
      initialize_char_type();
    }

    if(cur_ == end_)
    {
      token.type_ = CPP_Token::eof;
      token.line_ = line_;
      token.file_ = file_;
      return;
    }


    bool already_parsed = false;

    char c='@';

    // skip leading whitespace, comments, and #line directives

    if(first_char)
    {
      scan_till_first_nonblank();
	
      if(cur_ != end_ && *cur_ == '#')
	parse_preprocessor_directive();
    }

    while(cur_ != end_ &&
	  ( char_type_[*cur_] == ' '  ||
	    *cur_ == '\n' ||
	    *cur_ == '/'  ||
	    *cur_ == '\\'
	  )
	 )
    {
      c = *cur_;

      if(c == '\\')
      {
	if(cur_ == end_)    // handle line continuations
	  c = ' ';
	else
	{
	  ++cur_;
	  c = *cur_;
	}
      }

      if(c == '/')
      {
	
	++cur_;
	
	if(*cur_ == '/')
	{
	  // handle end of line comments
	
	  while(cur_ != end_ && *cur_ != '\n')
	    ++cur_;
	
	  if(cur_ != end_)
	  {
	    ++line_;
	    ++cur_;
	
	    scan_till_first_nonblank();
	
	    if(cur_ != end_ && *cur_ == '#')
	      parse_preprocessor_directive();
	
	  }
	
	  // now we should be at the beginning of the next line
	
	}
	else
	if( *cur_ == '*')
	{
	  // handle old style C comments
	
	  ++cur_;
	
	  while(cur_ != end_)
	  {
	    // scan for the pair, "*/"
	
	    c= *cur_;
	
	    if(c == '\n')
	    {
	      ++line_;
	    }
	
	    ++cur_;
	
	    if(cur_ == end_)
	      break;
	
	    if(c == '*' && *cur_ == '/')
	    {
	      ++cur_;
	      break;
	    }
	  }
	
	
	}
	else
	{
	  already_parsed = true;
	
	  break;
	}
	
      }
      else
      if( c == '\n')
      {
	++line_;
	


	if(cur_ != end_)
	{
	  ++cur_;

          scan_till_first_nonblank();
	
	  if(*cur_ == '#')
	    parse_preprocessor_directive();
	}
      }
      else
      {
	// handle whitespace
	
	while( cur_ != end_ && char_type_[*cur_] == ' ' )
	  ++cur_;
	
      }


    }


    // at this point we know we are not at the end_ of the stream, we can read and
    // classify the first character in the token


    if(cur_ == end_)
    {
      token.type_ = CPP_Token::eof;
      token.line_ = line_;
      token.file_ = file_;
      return;
    }

    if(!already_parsed)
    {
      c = *cur_;
      ++cur_;
    }

    char type = char_type_[c];

    token.text_ +=(c);

    token.line_ = line_;
    token.file_ = file_;

    // now, read any remaining characters in the token

    if(c == '\'')
    { // single character constants

       // the return value will contain the beginning and ending ' marks

       token.type_ = '\'';

       while(cur_ != end_)
       {
	 char secondary_character = *cur_;

	 if(secondary_character == '\\')
	 {
	   token.text_ +=('\\');
	   ++cur_;
	
	   if(cur_ != end_)
	   {
	     token.text_+=(*cur_);
	     ++cur_;
	   }
	   continue;
	 }

	 token.text_+=(secondary_character);
	 ++cur_;
	
	 if(secondary_character == '\n')
	 {
	   ++line_;
	   // no line number override should be occuring because
	   // this is technically a bug in the use of the language that
	   // I am trying to handle gracefully
	 }

	 if(secondary_character == '\'')
	 {
	   break;
	 }
       }

    } // single character constants
    else
    if(c == '"')
    { // string constant

      // the return value will contain the leading and trailing double quotes

      token.type_ = '"';

       while(cur_ != end_)
       {
	 char secondary_character = *cur_;

	 if(secondary_character == '\\')
	 {
	   token.text_+=('\\');
	   ++cur_;
	
	   if(cur_ != end_)
	   {
	     token.text_+=(*cur_);
	     ++cur_;
	   }
	   continue;
	 }

	 token.text_+=(secondary_character);
	 ++cur_;
	
	 if(secondary_character == '\n')
	   ++line_;
	   // no line number override should be occuring because
	   // this is technically a bug in the use of the language that
	   // I am trying to handle gracefully
	

	 if(secondary_character == '"')
	 {
	   break;
	 }
       }

    } // string constant
    else
    if( c == '.')
    {
      token.type_ = c;

      while(*cur_ == '.' && cur_ != end_)
      {
	token.type_ = CPP_Token::eli;
	token.text_ += c;
	c = *cur_;
	++cur_;
      }

      return;
    }
    else
    if( c == '(' || c == ')' || c == '[' || c == ']' || c == '{' ||
	c == '}'
      )
    {
      token.type_ = c;
      return;
    }
    else
    {
      bool is_number=false;

      if(c >= '0' && c <= '9')
      {
	is_number = true;
	token.type_ = CPP_Token::nmb;
      }
      else
      if( (c >= 'a' && c <= 'z') ||
	  (c >= 'A' && c <= 'Z') ||
	  (c == '_')
	)
      {
	token.type_ = CPP_Token::aln;
      }
      else
      if( c == '-' && *cur_ == '>')
      {
	token.type_ = CPP_Token::ptr;
	token.text_ += '>';
	++cur_;
	return;
      }

      while(cur_ != end_)
      { // main loop to aggregate characters of similar types

	char secondary_character = *cur_;

	while( secondary_character == '\\')
	{
	  // handle line continuations
	
	  ++cur_;
	
	  if(cur_ != end_)
	    secondary_character = *cur_;
	  else
	    secondary_character = ' ';
	
	  if(secondary_character == '\n')
	  {
	    ++cur_;
	    ++line_;
	
	    if(cur_ != end_)
	      secondary_character = *cur_;
	    else
	      secondary_character = ' ';
	  }
	
	}

	if(secondary_character == '.' && !is_number && type == 'A')
	  break;  // do not include '.' in names


	char secondary_type = char_type_[secondary_character];
	
	if(secondary_type != type)
	{
	  if( !  (secondary_character == '=' && type == '>')  )
	  {
	    break;
	  }
	
	}

	token.text_+=(secondary_character);

	++cur_;

      } // main loop to aggregate characters of similar types

      token.type_ = token_type(token.text_);

    }

  }

  int token_type(std::string const &s)

  /// Given an arbitrary string, determine what kind of token it is.
  /// The return value will be one of CPP_Token's types.  The token
  /// is assumed to be aligned at the beginning of the string with no
  /// leading blanks, tabs, etc.
  ///
  /// @return the type of token held in the input string
  /// @param s the input string to check on.

  {

    char c = s[0];

    int size = s.size();

    switch(c)
    {
    case '0': case '1': case '2': case '3': case '4': case '5':
    case '6': case '7': case '8': case '9':
      return CPP_Token::nmb;

    case '=':
      if(size == 2 && s[1] == '=')
	return CPP_Token::eql;
	
      break;

    case ':':
      if(size == 2 && s[1] == ':')
	return CPP_Token::der;
	
      break;

    case '&':
      if(size == 1)
	return '&';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '=')
	  return CPP_Token::ade;
	
	if(c == '&')
	  return CPP_Token::lan;
	
      }
      return CPP_Token::inv;

    case '|':
      if(size == 1)
	return '|';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '=')
	  return CPP_Token::ore;
	
	if(c == '|')
	  return CPP_Token::orr;
	
      }
      return CPP_Token::inv;

    case '!':
      if(size == 1)
	return '!';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '=')
	  return CPP_Token::neq;
	
      }
      return CPP_Token::inv;

    case '^':
      if(size == 1)
	return '^';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '=')
	  return CPP_Token::ceq;
	
      }
      return CPP_Token::inv;

    case '%':
      if(size == 1)
	return '%';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '=')
	  return CPP_Token::mde;
	
      }
      return CPP_Token::inv;

    case '/':
      if(size == 1)
	return '/';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '=')
	  return CPP_Token::dve;
	
      }
      return CPP_Token::inv;

    case '~':
      if(size == 1)
	return '~';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '=')
	  return CPP_Token::teq;
	
      }
      return CPP_Token::inv;

    case '*':
      if(size == 1)
	return '*';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '=')
	  return CPP_Token::tme;
	
      }
      return CPP_Token::inv;

    case '+':
      if(size == 1)
	return '+';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '+')
	  return CPP_Token::inc;
	
	if(c == '=')
	  return CPP_Token::ple;
	
      }
      return CPP_Token::inv;

    case '-':
      if(size == 1)
	return '-';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '-')
	  return CPP_Token::dec;
	
	if(c == '=')
	  return CPP_Token::mne;
	
      }
      return CPP_Token::inv;

    case '>':
      if(size == 1)
	return '>';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '>')
	  return CPP_Token::shr;
	
	if(c == '=')
	  return CPP_Token::geq;
	
      }
      else
      if(size == 3)
      {
	return CPP_Token::sre;
      }
      return CPP_Token::inv;

    case '<':
      if(size == 1)
	return '<';
      else
      if(size == 2)
      {
	c = s[1];
	
	if(c == '<')
	  return CPP_Token::shl;
	
	if(c == '=')
	  return CPP_Token::leq;
	
      }
      else
      if(size == 3)
      {
	return CPP_Token::sle;
      }
      return CPP_Token::inv;
	
    default:

      if( (c >= 'a' && c <= 'z') ||
	  (c >= 'A' && c <= 'Z') ||
	  c == '_'               ||
	  c == '$'
	)
	return CPP_Token::aln;
	
      break;
    }

    if(size == 1)
      return s[0];

    return CPP_Token::inv;

  }

private:

  void initialize_char_type()
    //
    /// initialize the array of character types so that the cpp_token function
    /// can combine tokens of similar types without having to do a lot of
    /// if then elses
    //
  {
    initialized_ = true;

    unsigned i;

    // first, give every character its own unique type

    for(i=0; i < sizeof(char_type_); ++i)  char_type_[i] = i;

    // next define classes of characters

    for(i=0; i < ' '; ++i)   char_type_[i] = 1;   // type 1 is an invalid character
    for(i=128; i < 256; ++i) char_type_[i] = 1;

    char_type_[' '] = ' ';  // a blank is a blank
    char_type_['\t']= ' ';
    char_type_['\r']= ' ';
    char_type_[0x0c]= ' ';  // \f not supported everywhere

    char_type_['\\']= '\\'; // this is a special case


    for(i='A'; i <= 'Z'; ++i) char_type_[i] = 'A';  // type 'A' is alphanumerics
    for(i='a'; i <= 'z'; ++i) char_type_[i] = 'A';  // underscore and '.' are also
    for(i='0'; i <= '9'; ++i) char_type_[i] = 'A';  // treated as alphanumeric -- you must
    char_type_['_'] = 'A';                          // split the returned tokens yourself
    char_type_['.'] = 'A';                          // if you don't want them treated this way.
    char_type_['$'] = 'A';

    char_type_['+'] = '=';  // type '=' refers to operators that can combine with '='
    char_type_['-'] = '=';
    char_type_['*'] = '=';
    char_type_['/'] = '=';
    char_type_['%'] = '=';
    char_type_['&'] = '=';
    char_type_['|'] = '=';
    char_type_['~'] = '=';
    char_type_['^'] = '=';
    char_type_['='] = '=';
    char_type_['<'] = '>';
    char_type_['>'] = '>';
    char_type_['!'] = '=';

  }


  void parse_preprocessor_directive()
    //
    /// parse any of the following:
    ///
    /// @li "#line number [name]"
    /// @li "# number [name]"
    /// @li "#pragma ..."
    /// @li "#define ..."
    ///
    /// If a preprocessor_handler_ is specified,
    /// use its function call operator method interpret the directive
    /// (other than line number directives which are consumed here)
    //
    //
  {
    // the stream should contain '# [line] number ["file"] [stackdepth]'

    while(cur_ != end_ && *cur_ == '#')
    {

      ++cur_;

      // skip white spaces before the 'line' keyword or the number

      while(    (cur_ != end_) 
            &&  (
                    (*cur_ == ' ') 
                 || (*cur_ == '\t')
                )
           )
	++cur_;

      if(cur_ == end_)
	return;

      if(*cur_ == 'l')
      {
	// assume this is a line keyword

	while(cur_ != end_ && *cur_ != ' ')
	  ++cur_;

	// skip white spaces after the 'line' keyword

	while(cur_ != end_ && *cur_ == ' ')
	  ++cur_;
	
	if(cur_ == end_)
	  return;
      }

      char c = *cur_;

      if(c < 0 || c > '9')
      {
	// this is a pragma or other ignorable pre-processor directive
	// handle line continuations using either \ or comments that start
	// in the macro body and don't complete on the same line
	
	std::string directive;       // could also be named 'accumulator'
	int         start_line=line_;

	char slash        = ' ';
	bool in_a_comment = false;

	while(cur_ != end_ && *cur_ != '\n')
	{
	  slash = *cur_;
	  ++cur_;
	
	  if(!in_a_comment)
	    directive += slash;
	
	  if(cur_ != end_)
	  {
	    if(*cur_ == '*' && slash == '/')
	      in_a_comment = true;
	    else
	    if(*cur_ == '/' && slash == '*')
	      in_a_comment = false;
	  }
	
	}
	
	if(cur_ != end_)
	{
	  ++cur_;  // skip the '\n'
	  ++line_; // keep the line_ variable correct

	  if(!in_a_comment)
	    directive += '\n';

	  scan_till_first_nonblank();
	
	  while(slash == '\\' || in_a_comment)
	  {
	    // the next line is a continuation line
	
	    while(cur_ != end_ && *cur_ != '\n')
	    {
	      slash = *cur_;
	      ++cur_;
	
	      if(!in_a_comment)
		directive += slash;
	
	      if(cur_ != end_)
	      {
		if(*cur_ == '*' && slash == '/')
		  in_a_comment = true;
		else
		if(*cur_ == '/' && slash == '*')
		  in_a_comment = false;
	      }
              else
              {
                in_a_comment=false;
              }
	
	    }
	
	    if(cur_ != end_)
	    {
	      ++cur_;  // skip the '\n'
	      ++line_; // keep the line_ variable correct
	
	      if(!in_a_comment)
		directive += '\n';

	      if(scan_till_first_nonblank())
		slash=' ';

	    }
            else
            {
             slash=' ';
             in_a_comment=false;
            }
	  }
	
	}

	if(preprocessor_handler_)
	  (*preprocessor_handler_)(directive, file_, start_line);

	continue;

      }

      // ok, so we should have 'number ["file"]'

      long line=0;

      while(cur_ != end_)
      {
	c = *cur_;

	if(c < '0' || c > '9')
	  break;
	
	++cur_;

	line *= 10;
	line += c - '0';
      }

      line_ = line;

      // so we have eaten the line number, now see if a file name
      // is specified

      // skip white spaces after the line number

      while(cur_ != end_ && *cur_ == ' ')
	++cur_;

      if(cur_ == end_)
	return;

      if(*cur_ == '\n')
      {
	++cur_;
        while(cur_ != end_ && char_type_[*cur_] == ' ')
         ++cur_;
	continue;
      }

      // ok we have a filename specified

      ++cur_;  // skip the leading "

      file_.resize(0);

      while(cur_ != end_ && (c = *cur_) != '"' && c != '\n' )
      {
	file_ += c;
	++cur_;
      }

      while(cur_ != end_ && *cur_ != '\n')
	++cur_;

      if(cur_ != end_)
      {
	++cur_;  // skip the \n
        while(cur_ != end_ && char_type_[*cur_] == ' ')
         ++cur_;
      }

    }

  }

  bool scan_till_first_nonblank()
    /// Skip characters in the stream until the first nonblank
    /// character is found.  Handle line number changes in
    /// blank sequences.
  {
    bool rv = false;

    for(;;)
    {
      // skip blanks and blank lines

      while(cur_ != end_ && char_type_[*cur_] == ' ')
	++cur_;
	
      if(cur_ == end_ || *cur_ != '\n')
	return rv;
	
      // *cur_ == '\n'

      ++cur_;
      ++line_;
      rv = true;

	
    }

  }

};

struct CPP_Token_Source
  ///
  /// A generic representation of a stream of CPP_Token.
  ///
  /// Derive from this and use an instance of CPP_Token_Stream to implement
  /// the operator function.  Execute the CPP_Token_Source objects as
  /// functions in order to parse the next token.  When the token
  /// stream is empty, the token type will be CPP_Token::eof.
{
  virtual void operator()(CPP_Token&) = 0;
    ///< User defined get next token function

  virtual ~CPP_Token_Source() {}
    ///< virtual destructor
};



class CPP_Stream_Token_Source
: public CPP_Token_Source
///
///  A source of CPP_Token's which comes from an istream.
/// 
///  A CPP_Stream_Token_Source is constructed from an extant stream -- it assumes
///  that the stream is as yet unread though that is not a requirement.  If you
///  do pass it a stream that is partially read, the line numbers associated
///  with tokens may be wrong.
//
{
public:

  CPP_Token_Stream<std::istreambuf_iterator<char> > *stream_;
    ///< stream from which to read the tokens

  CPP_Stream_Token_Source(std::istream &istr,
			  std::string filename="",
			  CPP_Token_Stream_Prep const* h=0)
  {
    ///< Construct a stream token source form a stream 
    /// -- optionally providing it with the name of the file and a pointer to a handler
    /// to take care of pre-processor directives.

     std::istreambuf_iterator<char> cur(istr);
     std::istreambuf_iterator<char> end;

     stream_ = new CPP_Token_Stream<std::istreambuf_iterator<char> >(cur,end,filename,h);
	
  }

  ~CPP_Stream_Token_Source()
    /// destructor
  {
    delete stream_;
  }


  void operator() (CPP_Token &token)
    /// Read the next token from the stream
  {
    (*stream_)(token);
  }

};

class CPP_File_Token_Source
: public CPP_Stream_Token_Source
//
///  A source of CPP_Token's which comes from a named file
//
{
  std::ifstream *input_file_; ///<  file stream from which to read tokens

public:

  CPP_File_Token_Source(char const *f = "", CPP_Token_Stream_Prep const* h=0)
  : CPP_Stream_Token_Source( *( input_file_ = new std::ifstream(f) ), f, h )
  {
    ///  Construct a token soruce from a char const *filename and a preprocessor directive
    ///  handler (optional).
  }

  CPP_File_Token_Source(std::string const &f, CPP_Token_Stream_Prep const* h=0)
  : CPP_Stream_Token_Source( *( input_file_ = new std::ifstream(f.c_str()) ), f, h )
  {
    ///  Construct a token soruce from a  string filename and a preprocessor directive
    ///   handler (optional).

  }

  ~CPP_File_Token_Source()
  {
    /// destructor

    delete input_file_;
  }

  operator bool() const { return bool(*input_file_); } ///<  check to see if file opened successfully

};


class CPP_String_Token_Source
: public CPP_Token_Source
{

  std::string copy_; ///<  copy of string being parsed

  CPP_Token_Stream<std::string::const_iterator> stream_; ///<  token stream pointing to this string

public:

  CPP_String_Token_Source(std::string s)
  : copy_(s),
    stream_(copy_.begin(), copy_.end())
  {
    ///
    /// Construct a token source (ie source of multiple tokens) from a string
    ///
    /// @param s is the string to convert into a sequence of tokens
  }

  void operator() (CPP_Token &token)
    /// read the next token from the string
    ///
    /// @param token is the token parsed from the string (one of many)
  {
    stream_(token);
  }

};


class CPP_Buffer_Token_Source
: public CPP_Token_Source
  ///
  /// A source of tokens extracted from a character array
  ///
{

  CPP_Token_Stream<char const *> stream_; ///< token source object

public:

  CPP_Buffer_Token_Source(char const *begin,
			  char const *end,
			  CPP_Token_Stream_Prep const *h=0)
  : stream_(begin, end,"",h)
  {
    ///
    /// Construct a token source from a pointer to the beginning and ending of
    /// a character buffer.
    ///
    /// @param begin the first character of the string of tokens
    /// @param end   the last character of the token source
    /// @param h     the preprocessor directive handler.
  }

  CPP_Buffer_Token_Source(char const *begin,
			  char const *end,
			  std::string f,
			  CPP_Token_Stream_Prep const *h=0)
  : stream_(begin, end,f,h)
  {
    ///
    /// Construct a token source from a character buffer but pretend it is
    /// from a file whose name is specified (and an optional preprocessor
    /// directive handler).
    ///
    /// @param begin the first character of the string of tokens
    /// @param end   the last character of the token source
    /// @param f     the name of the file to pretend to use
    /// @param h     the preprocessor directive handler.
  }


  void operator() (CPP_Token &token)
    /// Read the next token from the stream.
    ///
    /// @param token the destination of the parsed token
  {
    stream_(token);
  }

  char const * cur() const { return stream_.cur_; } ///<  current location in the input buffer
  char const * end() const { return stream_.end_; } ///<  end location in the input buffer



};

} // namespace cxxtls
#endif
