#ifndef CXXTLS_FMTSPEC_H_INCLUDED_P
#define CXXTLS_FMTSPEC_H_INCLUDED_P

#include <portable_io.h>

namespace cxxtls
{

struct fmtPad
   /// a unique data type that can serve as the specification of the leading
   /// pad character when formatting numbers.  The default is ' '.  When
   /// padding with leading zeros, you can construct this class like this:
   ///
   ///@begincode
   ///
   ///   fmtPad('0')
   ///
   ///@endcode
   ///
   /// This class exists primarily to be passed as a the last parameter of a fmtspec
   /// construction.  Yes, you can use other characters than '0', but won't actually
   /// work because the whole purpose of fmtPad is to populate the '0 field in
   /// the following sprintf call:
   ///
   ///@begincode
   ///
   ///   sprintf(buffer, "%-04d", 19);
   ///                      ^
   ///@endcode
   ///
   /// If the leadingPadChar_ member is ' ', then the above conversion specification
   /// will look like this:
   ///
   ///@begincode
   ///
   ///   sprintf(buffer, "%-4d", 19);
   ///                      ^
   ///
   ///@endcode
{
    char leadingPadChar_;

    fmtPad()
    :  leadingPadChar_(' ')
    {
    }

    fmtPad(char const &c)
    : leadingPadChar_(c)
    {
        if(c != ' ' && c != '0')
           throw *this;
    }
};


struct fmtspec
    /// A string formatting specification used by some of the lexical_cast template
    /// function overloaded signatures.  This class can also be used to construct 
    /// control strings directly usable in sprintf calls.  For example:
    ///
    ///@begincode
    ///
    ///    fmtspec tmp(-8,'u');
    ///
    ///    char buffer[40];
    ///
    ///    char outputBuffer[100];
    ///
    ///    snprintf(outputBuffer, 40, tmp.cs(buffer,0), short(-1));
    ///
    ///    // the above populates outputBuffer with "65535   "
    ///
    ///@endcode
    ///
    /// This class
    /// is provided to help you avoid manually calling sprintf() just to get a few 
    /// spaces or other formatting of a string to number conversion.  Instead, you
    /// can use cxxtls::lexical_cast() with an extra parameter and get the 
    /// benefits of type induction by the compiler.
    ///
    /// Like boost::lexical_cast, you can use the cxxtls::lexical_cast function like
    /// this:
    ///
    ///@begincode
    ///
    ///   std::string someVariable = cxxtls::lexical_cast<std::string>(19);
    ///
    ///@endcode
    ///
    /// That is, you can use it to simply convert some integer to a string (and
    /// viceversa if you reverse the type specifications).
    ///
    /// The fmtspec class exists to let you expand on this capability by adding
    /// formatting specifications tothe lexical_cast call.
    ///
    ///@begincode
    ///
    ///  std::string formattedVariable = 
    ///             cxxtls::lexical_cast<std::string>(19, 
    ///                                               cxxtls::fmtspec(4));
    ///@endcode
    ///
    /// In this example, instead of getting a string containing "19", you
    /// would get a string containing "  19".  That is, the string 19
    /// fit into a field of width 4. 
    /// 
    /// In general, fmtspec is a specification of a numeric or string format using the
    /// parameters to formatting that one sees in printf:
    ///
    /// @li width (including negative widths which mean left justified)
    /// @li precision (which means the maximum field width for strings and
    ///     which means digits to the right of the decimal point for floats).
    /// @li style which includes 'e', 'f', and 'g' for floating point numbers and
    ///     'o', 'h', 'b', and 'd' types for integers. 
    ///
    /// The style types have the following meaning:
    ///
    /// @li 'd' -- decimal
    /// @li 'x' -- hex
    /// @li 'o' -- octal
    /// @li 'u' -- unsigned
    ///
    /// In addition to mimicing the styles of sprintf when used with lexical_cast, 
    /// the fmtspec object is capable of
    /// generating the control string that you can then feed directly to sprintf(),
    /// or, more safely, snprinft(), to actually generate a real string from a number
    /// using normal tools.  (note that on MS Windows, snprintf is named _snprintf).
    ///
    /// Once you have constructed a fmtspec, you will use it like this:
    ///
    /// @li call preferredFormat(desiredDefaultFormatForYourDataType)
    /// @li call cs() to generate a constrol string for use by sprintf
    /// @li use it to call sprintf
    ///
    /// Note that the above styles apply to both int, long, and long long, you don't have
    /// different specifications for different data types.
    ///
    /// The fmtspec definition exists to assist in the more generic formatting
    /// of casts to std::string.  However, when converting to integers, the fmtspec
    /// object makes sense for specifying the string format of text to be parsed (oct,
    /// hex, decimal, etc.  When writing your own class overloads for lexical_cast,
    /// the use fmtspec is completely optional.
    ///
    /// Important working assumption:  It is not necessary to set any of the fields
    /// in a fmtspec constructor because they will default to values useful to the
    /// string conversions defined in this file.  This may seem strange since a 
    /// style spec would be an obvious requirement:  after all, you would seem to have
    /// tell the formatter that you want a number string conversion (for example), but
    /// in the lexical_cast() methods, we already know that.  Note that the lexical
    /// cast functions which use fmtspec will always call the preferredFormat method
    /// which sill set the style correctly if you don't set it yourself.  Thus, you
    /// are safe to only set the following:
    ///
    /// @li no members at all
    /// @li only the width
    /// @li only the width and precision
    /// @li all three
    ///
    /// The question then arises, if setting the style is so unnecessary why provide it
    /// at all?  Simple: there are some style options that you obviously need to set
    /// from time time:
    ///
    /// @li hex instead of decimal
    /// @li 'e' or 'f' format of floats instead of the default 'g'
    ///
{
  struct bad_format{};  // Missing style setting or other incompatible member settings.  
                        // Did you forget to set style?

  int width;
  int precision;
  char style;

  char buffer[40];

  fmtPad fmtPad_;      // the blank or zero used to add leading spaces/blanks to get the
                     // the desired field width.


  fmtspec()
  :  width(0)
  ,  precision(0)
  ,  style(0)
  ,  fmtPad_()
  {
     /// A truly empty format specification.  Probably not useful, but completeness
     /// is often a good thing to have.

     buffer[0]=0;
  }

  fmtspec(int w, int p, char const &s, fmtPad z=fmtPad())
  :  width(w)
  ,  precision(p)
  ,  style(s)
  ,  fmtPad_(z)
  {
      /// A fully user defined format specification
     buffer[0]=0;
  }

  fmtspec(int w, fmtPad z=fmtPad())
  :  width(w)
  ,  precision(0)
  ,  style(0)
  ,  fmtPad_(z)
  {   
     /// set the minimum width of the output fields, ignored on input
     buffer[0]=0;
     
  }

  fmtspec(int w, int p, fmtPad z=fmtPad())
  :  width(w)
  ,  precision(p)
  ,  style(0)
  {
    /// set the width and floating point precsion (for string formatting max
    /// length
  }

  fmtspec(char const &c, fmtPad z=fmtPad())
  : width(0)
  , precision(0)
  , style(c)
  {
     /// set only the desired output or input style
     buffer[0]=0;
  }

  fmtspec(int w, char const &c, fmtPad z=fmtPad())
  : width(w)
  , precision(0)
  , style(c)
  {
     /// set minimum field width and style
     buffer[0]=0;
  }

  void preferredStyle(char const &c)
  {
     /// chose a default style if it is not already
     /// set.

     if(!style)
        style = c;
  }

  void outputNumberBackwards(char *& rv, int number)
  {
     while(number)
     {
         int bottomDigit = number % 10;
         number /= 10;

         *(--rv) = '0' + bottomDigit;
     }


  }

  char const * cs(char cs[40], unsigned char size=0)
      /// Generate a formatspec for use by sprintf.  the
      /// cs parameter is the output buffer where the
      /// sprintf format string gets created.  The size parameter
      /// should be one of the following values:
      ///
      /// @li 0 -- means that a string or int or char is being
      ///          converted -- specifically, not a 'long' or
      ///          long long.
      /// @li 1 -- means that a long is being converted
      /// @li 2 -- means that a long long is being converted.
      ///
      /// The function returns a char const * into the buffer you
      /// have provided, so you use this function like this:
      ///
      ///@begincode
      ///
      ///   char csbuffer[40];
      ///
      ///   char outputBuffer[100];
      ///
      ///   fmtspec  s(...) ;  // initialize this somehow
      ///
      ///   sprintf( outputBuffer, s.cs(csbuffer), someVariable);
      ///
      ///   outputBuffer now holds someVariable converted into a string in outputBuffer
      ///
      ///@endcode
  {
     if(size > 2 || style == 0)  // did you forget to call preferredStyle() or set the style?
        throw bad_format();
     

     if(!   (   isIntegerStyle() 
                || isFloatStyle()
                || style == 's'
             )
       )
        throw bad_format();
        

     cs[39] = 0;
     
     char *rv = &cs[39];
     
     *(--rv) = style;  

                  
     if(isIntegerStyle() || isFloatStyle())
     {
         if(size == 1)
            *(--rv) = 'l';  // data will be long
         else
         if(size == 2)
         {
            *(--rv) = 'l';  
            *(--rv) = 'l';  // data will be long long
         }
         
     }

     if(precision)
     {
          outputNumberBackwards(rv, precision);
     
          *(--rv) = '.';  
     
     }
     
     if(width)
     {
          bool isNegative= width < 0;
     
          if(isNegative)
          {
             outputNumberBackwards(rv, 0 - width);

             if(fmtPad_.leadingPadChar_ != ' ')
               *(--rv) = fmtPad_.leadingPadChar_;
             
             *(--rv) = '-';
          }
          else
          {
            outputNumberBackwards(rv, width);

             if(fmtPad_.leadingPadChar_ != ' ')
               *(--rv) = fmtPad_.leadingPadChar_;

          }
     }
     else
     if(precision)
        *(--rv)='0';
     
     *(--rv) = '%';
     
     return rv;
     
  }

  bool isFloatStyle() const
  {
     return     style == 'f' 
            ||  style == 'e'
            ||  style == 'g'
            ;
  }

  bool isIntegerStyle() const
  {
     return     style == 'd' 
            ||  style == 'o'
            ||  style == 'x'
            ||  style == 'u'
            ;
  }


};

    template<class Integer>
    void signedInt2String(char buffer[40], Integer const &t, char preferred, fmtspec s=fmtspec())
        /// convert an integer into a string and store it in a character buffer.
    {
       long long tmp = t;

       s.preferredStyle(preferred);

       char cs[40];

       if(s.width > 35)
          throw fmtspec::bad_format();  
              // jeez guys, figure out a way to handle your string width
              // without expecting built in conversions do deal with any
              // random crap you throw at it.  35 characters are enough
              // for any reasonable basic conversion utility.


       snprintf(buffer, 40, s.cs(cs,2), tmp);  // the 2 means long long

       buffer[39] = 0;
    
    }

}
#endif
