#ifndef cxxtls_thread_h_included_p
#define cxxtls_thread_h_included_p
//
// Copyright 2009-2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <deque>
#include <iostream>
#include <iomanip>

#ifdef _MSC_VER
#  include <windows.h>
#  define  LL_MUTEX_IMPL  CRITICAL_SECTION     /* LOW LEVEL MUTEX IMPLEMENTATION */
#else
#  include <pthread.h>
#  define  LL_MUTEX_IMPL  pthread_mutex_t      /* LOW LEVEL MUTEX IMPLEMENTATION */
#endif  

namespace cxxtls
{

class Mutex
   ///
   /// A Mutex is an object which multiple threads can use to synchronize
   /// access to shared resource.  One thread can "lock" the mutex, which
   /// tells the other threads that they must wait to access the shared
   /// resource -- or try again later.  When the owning thread releases the
   /// lock, if there are any other threads waiting for a lock, one of them
   /// will be selected (seemingly at random) to receive the lock.  Only
   /// the thread currently owning the lock should modify the shared resources.
   ///
   /// Mutex objects are meant to be locked and unlocked primarily using
   /// the helper class, ScopedMutexLock.   Before using a mutex directly,
   /// consider using both a Mutex and a ScopedMutexLock.  See it's documentation.
   ///
   /// Note that it is @b strictly @b forbidden for @b one @b thread @b to @b lock @b a @b mutex and
   /// for a @b different @b thread @b to @b unlock @b it.  Basically, the definition
   /// of mutexes makes the unlock operation not thread safe.  This makes it
   /// fundamentally difficult, using only mutexes, to communicate data between threads.
   /// The ConditionVariable class, below, is used to communicate state across threads.
   /// ConditionVariable relies on ScopedMutexLock objects and a Mutex, as well as 
   /// operating system provided mutexes to do its job.
   ///
   /// @note Non-Recursive Mutexes
   ///
   /// @warning  Mutexes may or may not be recursive.  On @b linux/unix they are
   ///           @b not, but on @b windows they @b are.  
   ///
   /// A recursive mutex is one in which the same thread can lock the same mutex multiple
   /// times, safely.  This is helpful in that it makes your code easier to write, because
   /// you don't have to worry that you might call functions in your own thread that blindly
   /// lock the mutex -- and thus hang your thread.  But, it also confuses the issue of proper
   /// sequencing on the mutex to allow this.  So, the official way that this class, Mutex,
   /// works is @b supposed @b to @b be non-recursive, but as mentioned before, on windows, the
   /// CRITICAL_SECTION objects that implement Mutex are by nature recursive.  Unfortunately,
   /// supporting recursive mutexes in the posix threads environment is somewhat expensive, so
   /// recursive mutexes are not yet supported here.  BOOST threads support recursive mutexes,
   /// so it must be possible.  There implementation could be used for guidance in designing
   /// one for cxxtls.
   ///
   ///@code
   ///
   ///  ... 
   /// 
   ///   main()
   ///   {
   ///       Mutex::enableLocks();  // must be called once at the beginning of the program
   ///                              // and never again
   ///   }
   ///
   ///   void someFunctionCalledByAThread()
   ///   {
   ///      static Mutex mutex;
   ///
   ///      ScopedMutexLock lock(mutex); // NOT STATIC!
   ///
   ///      // Only one thread can execute this code at a time
   ///
   ///   }  // mutex is unlocked here, and other threads can continue
   ///
   ///@endcode
   /// 
{
   LL_MUTEX_IMPL mutex_;


   // NOTE:  do not add a locked_ flag here.  You have to make a function call on
   // the mutex to determine if it is locked, you can't keep track of the lock flag
   // outside of the operating system.

public:


   LL_MUTEX_IMPL& mutex()
       ///  @return a pointer to the actual LL_MUTEX_IMPL object used by this
       ///  this Mutex.
   {
       return mutex_;
   }

   Mutex();

   ~Mutex();


   void lock();
      ///< Attempt to lock this mutex.  Hang until you are able to or
      /// until a fatal error occurs.  Throws an exception on failure

   void unlock();
       ///<  Unlock a locked mutex.  Throws an exception if not locked or other failure.

   bool tryLock();
       ///<  In a non-blocking manner, attempt to lock the mutex.  If the attempt was 
       ///   successful, return true.  Otherwise return false.

   struct LockFailureException{};          ///< could not lock the mutex
   struct UnlockFailureException{};        ///< could not unlock the mutex
   struct AttributeInitFailureException{}; ///< could not construct default atts
   struct InitFailureException{};          ///< could not initialize a mutex
   struct DestroyFailureException{};       ///< destructor error on mutex
   struct ConditionVariableSyncError{};    ///< some kind of logic bug 
   struct ConditionVariableWaitFailure{};  ///< error waiting for mutex
   struct ConditionVariableNotifyFailure{};///< error calling pthread_cond_signal

};


class ScopedMutexLock
    ///
    ///  An object that is used to lock a mutex, and whose construction
    ///  will not complete until the lock has been granted.
    ///
    ///  The locking is meant to be non-recursive, though on windows, it is because
    ///  that is how CRITICAL_SECTIONs work.  Beware when programming of the
    ///  possibility that your code will work on windows but hang on everything else.
    ///  See class Mutex, for a discussion on recursive mutexes.
    ///
    ///  A ScopedMutexLock object is constructed on the stack in a thread and
    ///  is passed a reference to a Mutex object which should exist in 
    ///  static memory or in heap.
    ///
    ///  Always give ScopedMutexLock objects a name to prevent the compiler from
    ///  destructing them immediately after construction.  For example:
    ///
    ///@code
    ///
    ///   ScopedMutexLock  locker(someMutexObject);
    ///
    ///@endcode
    ///
    ///  ScopedMutexLocks should in general not be static variables or in heap
    ///  packets -- the mutexes should be static or heap resident, but
    ///  locks generally serve no purpose when used in that way.
    ///
    ///
    ///  Members lock() and unlock() give you instantaneous control over the lock
    ///  or unlock state.  Note that the destructor can be safely called in either
    ///  the locked or unlocked state.  However, you cannot double lock or double
    ///  unlock a mutex without throwing an exception or creating some other kind
    ///  of error.
    ///
    ///  @note  The destructor does not throw an exception if the mutex has
    ///         been destructed by some other process.
{
    Mutex *mutex_;  ///<  The mutex locked by this scoped lock

    bool locked_;   ///<  A flag indicating that this ScopedMutexLock has in fact
                    ///   locked the mutex.
                                                              
public:

    ScopedMutexLock(Mutex &m)
    : mutex_(&m)
    {
        lock();
    }

    ~ScopedMutexLock()    // destructors are not supposed to throw for any reason
    {
        try
        {
            if(locked_)
                 mutex_->unlock();
        }
        catch(...)
        {
            return;
        }
    }

    void lock()
       /// a hard lock on the controlled mutex.  Not recursive or reentrant.
    {
       mutex_->lock();
       locked_ = true;
    }
    
    void unlock()
       /// unlock the mutex controlled by this ScopedMutexLock.  It is bug
       /// to unlock someone else's lock.
    {
        mutex_->unlock();
        locked_ = false;
    }
    
    Mutex *mutex() 
       /// @return a pointer to the mutex that can be controlled by this ScopedMutexLock
    {
       return mutex_;
    }
    
    bool isLocked() const
        ///  @return true if this ScopedMutexLock has in fact locked the mutex
        ///
        ///  This function is not good for anything but debugging
    {
       return locked_;
    }

};

class ConditionVariable
    ///  A ConditionVariable allows you to safely unlock then relock a Mutex from 
    ///  different threads.  Mutexes are meant for unlocking by a single thread
    ///  and if multiple threads try to unlock them, various problems can result.  A condition
    ///  variable is the basic mechanism for safely communicating information between
    ///  threads.  It does this by making the unlock process thread safe.
    ///
    ///  A condition variable is always associated with a mutex.  You use the condition
    ///  variable to make your thread wait until signaled by some another thread.  While
    ///  you are waiting, the mutex is unlocked.  Other threads can lock it safely if they
    ///  are also using the same condition variable.
    ///
    ///  To do this, call the wait() method giving it a reference to a ScopedMutexLock
    ///  which is in the locked state.  The wait() function will temporarily unlock
    ///  the ScopedMutexLock so as to allow some other thread to lock it and will wait
    ///  for it to become available for re-locking.  When said thread then unlocks said 
    ///  mutex, the wait() function will relock the thread and return.  This gives a
    ///  safe way for the a master thread to control the behavior of worker threads.
    ///
    ///  See class ThreadQueue<> below.  It demonstrates usage of a ConditionVariable.
    ///
    ///  Here is the basic usage scenario -- assuming you can call this "basic".
    ///
    ///@code
    ///
    ///   Thread 0:  Defines a variable of type Mutex.
    ///
    ///   Thread 1:  Uses ScopedMutexLock to lock the mutex.  This is the master
    ///              thread in control of the situation.
    ///
    ///              This thread then defines a ConditionVariable and calls its
    ///              wait() method given the above ScopedMutexLock.  This unlocks
    ///              the mutex and allows other threads to lock it.
    ///
    ///   Thread 2-N:   Uses ScopedMutexLock to lock the mutex for the purpose
    ///                 of changing some data guarded by the mutex.  Once the
    ///                 lock is granted, the one thread owning the lock can
    ///                 change the data.  It must then call the condition variable's
    ///                 notifyOne() method to tell the controlling thread that data
    ///                 is now available.  The destructor for the ScopedMutexLock
    ///                 will release the lock on the mutex and the condition variable's
    ///                 wait method will return in the controlling thread -- with the
    ///                 mutex being locked against any other threads being involved.
    ///
    ///   Thread 1:  When any thread (2-N) unlocks the mutex, if the condition variable's
    ///              notifyOne() method has been invoked, control immediately
    ///              returns to Thread 1 and the wait() method returns with the mutex
    ///              in a locked state.  The controlling thread will then consume the data
    ///              or interpret any state changes guarded by the mutex.  It should then
    ///              call the conditon variable's wait() method to allow other threads to 
    ///              supply data.
    ///
    ///
    /// Note that more complex scenarios are possible.  Instead of calling notifyOne(), the
    /// notifyAll() function could be called.   This would allow multiple threads to act as
    /// the master thread.
    ///
    ///
    ///@endcode
{
#   ifndef _MSC_VER
       pthread_cond_t   condition_;
#   else
       CONDITION_VARIABLE condition_;
#   endif

public:

    ConditionVariable();
    
    void wait(ScopedMutexLock &lock);
        // wait forever until a call to either notifyOne() or notifyAll() is made by some other
        // thread.
        //
        // While waiting, unlock the mutex so that other threads can grab a lock
        // then call this function on THIS object to signal us tht new data
        // is available.

    
    void notifyOne();
       ///<  Signal to some other thread that we are done changing data that that both
       ///  this and that other thread are regulating access using *this.
    

    void notifyAll();
       ///  Signal to all other threads which are @b wait() -ing on this condition variable
       ///  that we are done changing data and that they can safely consume the data.
       ///
       ///  Note that all waiting threads will be synchronized by their individual attempts
       ///  to lock the mutex that has been passed as a parameter in their calls to @b wait()
       ///  or to @b waitUntil().
       ///
       ///  Note that only one will receive the lock at a time.  If the first thread consumes
       ///  all the data, then there is really no reason have multiple waiters.  So if you
       ///  have slow moving threads that consume the data, you might want to have multiple
       ///  waiters, each of which only grabs part of the data and processes it before
       ///  calling wait() again.
       ///
       ///
       ///  @note  This function can be called by any thread.  It does not requre that
       ///         that the caller actually own the mutex lock that is presumably held
       ///         by a wait() or waitUntil() call.  @b But doing so will make it very
       ///         hard to control the program flow.  @b so @b basically: it is a bad
       ///         idea to do that.

    
    ~ConditionVariable();

};

template<class T>
class ThreadQueue
    ///  A thread safe input or output queue.
    ///
    ///  This queue is safe for multiple writers and multiple readers.  
    ///
    ///  By default, the queue assumes a single reader thread will be consuming
    ///  data, but if you wish to have multiple simultaneous consumers, you can,
    ///  when constructing the queue, pass a @b true value for the parameter, notifyMultiple.
    ///  The default value of this parameter is false.  Note that you can still have
    ///  multiple readers even with this parameter set to false.  The difference is that 
    ///  without the notifyMultiple flag being true, only 1 reader will be awakened at a time.
    ///  This gives you some control over parallelism even if you have multiple reader threads
    ///  -- but it is a poor substitue for designing the input data to enforce the ordering
    ///  of command execution.
    ///
    ///  This kind of queue is very helpful when writing threads which 
    ///  must read commands and execute them.  Multiple writer threads can inject
    ///  commands into the queue, and multiple reader threads can consume from the
    ///  queue and execute the commands.  Not that there positively no ordering
    ///  guarantee on the execution of commands if you have multiple readers and
    ///  multiple writers.  The content of the commands in the queue is forced to
    ///  include any ordering information that might be required.
    ///
    ///  Here is the basic use case for a ThreadQueue:
    ///
    ///  @li  Thread 1 uses the "push()" method to push objects into 
    ///       thread 2's input queue.
    ///
    ///  @li  Thread 2 repeatedly calls the waitPop() method to wait for
    ///       a command and execute it.   You should check the return code
    ///       from waitPop() as described below.
    ///
    ///  @li  Alternatively, Thread 2 can use one of the following other
    ///       means of getting commands out of the queue:
    ///       <blockquote>
    ///          <on>
    ///              <li> tryPop() -- which returns true and one T object if there
    ///                   is data waiting in the queue.   Otherwise, it returns false
    ///                   and no data.
    ///              <li> trySwapAll() -- which returns true and swaps the entire
    ///                   queue with a local vector of T objects if there are any.
    ///          </on>
    ///       </blockquote>
    ///
    /// The waitPop() member function returns true to indicate that data has
    /// arrived.  If the queue has not been terminated by some other thread,
    /// then waitPop() will wait forever until data arrives.  If some other
    /// thread uses the terminateInput() method on the queue, then all outstanding
    /// waitPop() calls will return immediately and will return false to indicate
    /// that an end of stream case has occured.
    ///
    /// Note that there is an empty() function to tell the caller if the
    /// the queue is empty, but since this can change instantaneously this
    /// function is only useful for debugging.
    ///
    /// The tryPop() method lets you check for data in the queueu and 
    /// return immediately if there is none.  This is useful for code that 
    /// has timeouts in it. 
    ///
    /// @b But @b remember, if you use the empty() or trypPop() methods, your
    /// application should include a call to nanosleep() to prevent your
    /// loop from burning up all the resources of the CPU on which it is 
    /// running.
    ///
    /// If input on the queue should be terminated, call member terminateInput()
    /// and you can call the function inputTerminate() to determine if this has
    /// happend -- or you can know that this has happened by noticing that waitPop()
    /// has returned false.  It only returns false if input has been terminated.

{
    std::deque<T>      data_;
    mutable Mutex      mutex_;
    ConditionVariable  cond_;
    bool               inputTerminated_;  ///<  the queue should be stopped asap
    bool               notifyAll_;

public:

    ThreadQueue(bool notifyMultiple = false)
    :  inputTerminated_(false)
    ,  notifyAll_(false)
        ///  Construct a thread safe queue.  By default, no parameters are required.  When a queue is constructed
        ///  in this way, it behaves as if there is only one consuming thread.  To target your queue for multiple
        ///  simultaneous consumers, then set the constructor parameter, notifyMultiple, to true on construction.
        ///
        ///  @note Multiple readers can still read from the queue even if the notifyMultiple is false, but in this
        ///  case, only one reader will be awakened when new data is pushed into the queue.
    {
    }


    ~ThreadQueue()
    {
       ScopedMutexLock lock(mutex_);
       data_.clear();
       inputTerminated_ = true;
    }


    void terminateInput()
        ///  signal a permanent end to operations on this queue.  This function
        ///  should only be called in preperation for thread cancellation or program
        ///  exit.
    {
       ScopedMutexLock lock(mutex_);
       inputTerminated_ = true;
       data_.clear();
       cond_.notifyAll();   // all is ok here because we are on the way out and there is no day
                            // for the awakened threads to process anyway.
    }

    bool inputTerminated() const 
        ///  return true if this queue's terminateInput() function has been called.
        ///  
    {
       return inputTerminated_;
    }

    void push(T const& t)
        /// Insert a new item into the queue and notify 1 or more of the waiting
        /// threads that the data is available.
        ///
        /// @note  If the terminateInput() function has been called, this function
        /// will do nothing.
    {
        ScopedMutexLock lock(mutex_);   // wait until write privilege is obtained

        if(inputTerminated_)
           return;  // accept no more data

        data_.push_back(t);             // save the item in the queue

        if(notifyAll_)
           cond_.notifyAll();
        else
            cond_.notifyOne();          // notify the consumer that data is available
    }

    bool empty() const
       ///  @return true if the queue is empty or if the terminateInput() function has been called.
    {
        ScopedMutexLock lock(mutex_);

        if(inputTerminated_)
           return true;

        return data_.empty();
    }

    bool tryPop(T& rv)
        /// @return true if data is avaialable.  Note that if the terminateInput() function has been
        /// called, this function will return false.  It will also return false if there is not
        /// data to be had.  If tryPop() returns false, you should call inputTerminated() to properly
        /// understand the situation.
    {
        ScopedMutexLock lock(mutex_);

        if(data_.empty())
        {
            return false;
        }
        
        rv = data_.front();
        
        data_.pop_front();
        
        return true;
    }

    void trySwapAll(std::deque<T> & rv)
        ///  Swap an existing (usually empty) deque of T objects
        ///  with the current contents of the data_ buffer.
        ///
        ///  Do so in a thread safe manner.
        ///
        ///  Generally this function is used to just empty the whole
        ///  container at once.  It can be used populate the queue
        ///  with data but this is an iffy idea.
        ///
        ///  The function can't fail, but may return you an empty
        ///  container.
        ///
        ///  Note that if input is terminated, this function will
        ///  discard any data that you swap into the queue.
    {
        ScopedMutexLock lock(mutex_);

        rv.swap(data_);

        if(inputTerminated_)
           data_.clear();
        
    }

    bool waitPop(T& rv)
        /// wait until data becomes available in the queue or until input is
        /// terminated.  Return the data if input is not terminated.  
        /// @return true if the data is available -- which it will be unless
        /// input is terminated.  If false is returned, interpret this as a
        /// termination state and do not make the call again.
    {
        ScopedMutexLock lock(mutex_);

        // we have write priviledge -- if there is data, go ahead and consume
        // one item, otherwise, let someone write something into the queue then
        // consume that

        while(!inputTerminated_ && data_.empty())
        {
            cond_.wait(lock); // restart the wait if the queue is empty because
                              // this function should only return if data is available.
        }

        if(inputTerminated_)
            return false;
        
        rv = data_.front();
        data_.pop_front();

        return true;
    }

};


/// Macro, CXXTLS_DECLARE_THREAD_FUNCTION, Applications can use to define thread functions in a portable manner.  That is,
/// use this macro to define thread functions rather than writing the code yourself.


#ifndef _MSC_VER

    #define CXXTLS_DECLARE_THREAD_FUNCTION(threadName,ParmName)   void *threadName(void *ParmName)
        ///<  Declare a function whose name can be used as a thread function.

    typedef void *(*ThreadFunction)(void*);                 ///< A pointer to a function which a thread can
                                                            ///  actually executed.
                                                       
    typedef pthread_t ThreadId;                             ///< A number as a handle to a thread.

    
#else

    #define CXXTLS_DECLARE_THREAD_FUNCTION(threadName,ParmName)  DWORD WINAPI threadName(void *ParmName)

    typedef LPTHREAD_START_ROUTINE ThreadFunction;    
                                                      
    typedef HANDLE ThreadId;                          

#endif


#ifdef _MSC_VER
ThreadId createThread(ThreadFunction function, void *threadData);  // on windows, you can't have a non-joinable thread
#else
ThreadId createThread(ThreadFunction function, void *threadData, bool joinable=true); 
#endif
    ///<  Creates a thread given a specified thread function pointer and
    ///   a specific piece of data to work on.  Also, depending on the @b joinable flag, specify
    ///   that this flag must be joined -- or not joined.  
    ///
    ///   A joinable thread that is not joined will hang the application if it attempts to exit.
    ///
    ///   On windows, all threads @b must @b be @b joinable.  The createThread function signature does not permit
    ///   non-joinable threads on windows.  Common parameters are as follows:
    ///
    ///   @param function is a pointer to the thread handling function which must be defined as is described below.
    ///   @param threadData is a pointer to memory which holds the parameters for this thread.  You will usually want
    ///          to create this memory with operator new and let your thread function delete it when it is done.
    ///          Use std::auto_ptr for saftey if you need to.
    ///   
    ///   On unix and linux:
    ///   Non-joinable threads, however, cannot be reliably expected to terminate in any given
    ///   time frame.  But, they do eliminate the need to keep track of them.  If you need to know
    ///   that a thread has actually finished, you should use joinable.  If you want to fire and 
    ///   forget, then joinable should be set to false.  Don't try to join a non-joinable thread.
    ///
    ///   @return 0 to indicate failure.
    ///
    ///   Normally, when creating threads, you want to have some class object be operated on
    ///   by thread.  Thus, you will want to do something like
    ///
    ///   @b Declaring @b Thread @b Functions
    ///
    ///   Functions which are to be passed to the createThread function should be declared / defined
    ///   using the macro: CXXTLS_DECLARE_THREAD_FUNCTION() using one of the following approaches:
    ///
    ///   Within the body of a class:
    ///@code
    ///
    ///   class  SomeClass 
    ///   {
    ///      @b static CXXTLS_DECLARE_THREAD_FUNCTION(funcName, funcParmName);   // class members must be STATIC!
    ///   };
    ///
    ///@endcode
    ///
    ///   Outside the body of class, use one of the following declarations/definitions:
    ///
    ///@code
    ///   
    ///    // Forward declarations
    ///     
    ///    extern CXXTLS_DECLARE_THREAD_FUNCTION(func_1_Name, func_1_ParmName);
    ///
    ///    static CXXTLS_DECLARE_THREAD_FUNCTION(func_2_Name, func_2_ParmName);
    ///
    ///    CXXTLS_DECLARE_THREAD_FUNCTION(func_3_Name, func_3_ParmName);
    ///
    ///    CXXTLS_DECLARE_THREAD_FUNCTION(func_4_Name, func_5_ParmName)
    ///    {
    ///      ...
    ///    }
    ///
    ///    static CXXTLS_DECLARE_THREAD_FUNCTION(func_5_Name, func_5_ParmName)
    ///    {
    ///      ...
    ///    }
    ///   
    ///  @note The parameter passed into a thread function is always a @b void @b * which you must
    ///        cast into some other data type for use.
    ///
    ///  @note Further, in many useful situations, you may feel a need to new up an object and pass the
    ///        address of it to your thread function -- @b which @b will @b have @b to @b delete @b it
    ///        at the end of the thread processing.
    ///
    ///@endcode
    ///
    ///  Note:  
    ///
    ///@code
    ///
    ///   class ThreadDataClass
    ///   {
    ///   public:
    ///
    ///       void realThreadFunction() { ... }  ///&lt; std c++ code to run the thread on this
    ///                                          ///  this object
    ///
    ///       static void threadInterfaceFunction(void *d)
    ///          /// function callable by createThread() give this as the data parameter
    ///       {
    ///            ThreadDataClass *me = reinterpret_cast<ThreadDataClass*>(d);
    ///
    ///            me->realThreadFunction();
    ///
    ///            delete me;  // otherwise, join will leak memory
    ///       }
    ///
    ///   };
    ///
    ///   ThreadId createNeededThread()
    ///   {
    ///       return createThread(  &ThreadDataClass::threadInterfaceFunction,  new ThreaDataClass);
    ///   }
    ///
    ///   void WaitForThreadtoComplete(ThreadId id)
    ///   {
    ///       joinThread(id);  // assumes that the thread cleaned up memory after itself.
    ///   }
    ///
    ///@endcode
    ///

bool joinThread(ThreadId id);
    ///
    ///  Join and destroy a specified thread.  Do not join a thread created with the joinable
    ///  flag set to false.
    /// 
    ///  @return false on failure.
    ///
    ///  Note:  a thread can't join itself -- if you try, the application will most likely hang.
    ///

enum ThreadTraceLevel
    ///  The amount of program trace information desired.
    ///
    ///  Program trace information is created using any of the following macros:
    ///
    ///  @li @ref CXXTLS_DEBUG
    ///  @li @ref CXXTLS_LOG
    ///  @li @ref CXXTLS_DEBUGNONL
    ///  @li @ref CXXTLS_LOGNONL
    ///  @li @ref CXXTLS_HEAVY_LOG
    ///  @li @ref CXXTLS_HEAVY_LOGNONL
    ///
    ///  The default program trace is TT_LOG.
    ///
    ///  The program trace level is controlled using CXXTLS_SET_PROGRAM_TRACE_LEVEL(t) where
    ///  t is one of members of this enumeration.  Higher numbers mean more volume of trace
    ///  info.  
{
    TT_NONE,      ///<  No trace output at all
    TT_LOG,       ///<  Normal user visible program trace output
    TT_DEBUG,     ///<  Progrmmer trace to aid in debugging
    TT_HEAVYDEBUG ///<  Programmer trace too voluminous for normal operations
};


extern Mutex coutLock;              ///< a mutex lock that guards stdout
extern ThreadTraceLevel coutLevel;  ///< Current trace level, defaults to TT_LOG.  Use
                                    ///  CXXTLS_SET_PROGRAM_TRACE_LEVEL(t) to change.


}; // namespace cxxtls


///
///  Macro, CXXTLS_COUT_LOCK(), declares a local variable named in the parameter which locks
///  multi-threaded output to stdout.  This can be used in your own code temporarily lock cout
///  from other multithreaded uses.  For example:
///
///@code
///
///     int c;
///     while( 0 > ( c = fgetc(file) ) )
///     {
///         CXXTLS_COUT_LOCK(myTempName);
///         cout << c;
///     }
///
///@endcode
///
///  Note that this variable does a mutex lock, so you might want to just save up characters till
///  you hit a new line then call CXXTLS_LOG() on the whole string.  As this is not always possible,
///  the above approach may be required.
///
#define CXXTLS_COUT_LOCK(var)   cxxtls::ScopedMutexLock  var(cxxtls::coutLock)

///
///  Macro, CXXTLS_SET_PROGRAM_TRACE_LEVEL(t), lets you set the current trace level.  
///  @see ThreadTraceLevel, for details
///
#define CXXTLS_SET_PROGRAM_TRACE_LEVEL(t) \
    {\
        CXXTLS_COUT_LOCK(var) ; \
        cxxtls::coutLevel = t; \
    }

///  Macro, CXXTLS_DEBUG(), is a thread safe debug console logging macro that produces
///  a line prefix for each message.  Messages are sent to cout and and endl is
///  automatically added.  All lines are prefixed with the file name and line number
///  of the source code point where the message is produced.
///
///  @param e  The expression to be printed.  Basically e can be anything that you 
///            would normally put in cout statement.
///
///
#define CXXTLS_DEBUG(e) \
      { \
          CXXTLS_COUT_LOCK(debugLogLock); \
          if(cxxtls::coutLevel >= cxxtls::TT_DEBUG) \
          std::cout << __FILE__ << ':' << std::dec << __LINE__ << ":  " << e << std::endl; \
      }

///  Macro, CXXTLS_DEBUGNONL(), is the same as CXXTLS_DEBUG() except that it does no
///  append a trailing newline to the message -- this is to account for messages that
///  already have a trailing newline.
///
#define CXXTLS_DEBUGNONL(e) \
      { \
          CXXTLS_COUT_LOCK(debugLogLock); \
          if(cxxtls::coutLevel >= cxxtls::TT_DEBUG) \
          std::cout << __FILE__ << ':' << __LINE__ << ":  " << e; \
      }

#define CXXTLS_HEAVYDEBUG(e) \
      { \
          CXXTLS_COUT_LOCK(debugLogLock); \
          if(cxxtls::coutLevel >= cxxtls::TT_HEAVYDEBUG) \
          std::cout << __FILE__ << ':' << __LINE__ << ":  " << e << std::endl; \
      }

///  Macro, CXXTLS_DEBUGNONL(), is the same as CXXTLS_DEBUG() except that it does no
///  append a trailing newline to the message -- this is to account for messages that
///  already have a trailing newline.
///
#define CXXTLS_HEAVYDEBUGNONL(e) \
      { \
          CXXTLS_COUT_LOCK(debugLogLock); \
          if(cxxtls::coutLevel >= cxxtls::TT_HEAVYDEBUG) \
          std::cout << __FILE__ << ':' << __LINE__ << ":  " << e; \
      }

///  Macro, CXXTLS_LOG(), is a thread safe console logging macro.
///  Messages are sent to cout and and endl is
///  automatically added.  
///
///  @param e  The expression to be printed.  Basically e can be anything that you 
///            would normally put in cout statement.
///
///  Unlike @ref CXXTLS_DEBUG(), this macro does NOT prefix lines with the source file
///  and line number.  So this macro can be used to produce regular console program output
///  in a threadsafe manner.
///
#define CXXTLS_LOG(e) \
      { \
          CXXTLS_COUT_LOCK(normalLogLock); \
          if(cxxtls::coutLevel >= cxxtls::TT_LOG) \
          std::cout << e << std::endl; \
      }

///
///   Macro CXXTLS_LOGNONL() is the same as CXXTLS_LOG() but it does not add
///   a new line to the message.
///
#define CXXTLS_LOGNONL(e) \
      { \
          CXXTLS_COUT_LOCK(anyOldNameYouLike); \
          if(cxxtls::coutLevel >= cxxtls::TT_LOG) \
          std::cout << e ; \
      }


#endif // cxxtls_thread_h_included_p
